// lasrc/rgb.c
// (C) 1998 lars brinkhoff

#include <stdio.h>
#include <stdlib.h>

#include "lasrc.h"
#include "image_out_device.h"

struct device
{
  Image_out_device dev;
  unsigned char *data;
  FILE *f;
};

static void
pixel_rgb (Image_out_device *device_, int x, int y,
	   double r, double g, double b)
{
  struct device *device = (struct device *)device_;

  device->data[3 * (y * device->dev.width + x) + 0] = (int)(255 * r + 0.5);
  device->data[3 * (y * device->dev.width + x) + 1] = (int)(255 * g + 0.5);
  device->data[3 * (y * device->dev.width + x) + 2] = (int)(255 * b + 0.5);
}

static void
wait_rgb (Image_out_device *device)
{
}

static void
close_rgb (Image_out_device *device_)
{
  struct device *device = (struct device *)device_;

  fwrite (device->data, 3, device->dev.width * device->dev.height,
	  device->f);
  fclose (device->f);
  free (device->data);
  free (device);
}

Image_out_device *
create_rgb_image_out_device (char *name, int width, int height, int scale,
			       double gamma)
{
  struct device *device;
  
  device = malloc (sizeof (struct device));
  if (device == 0)
    {
      fprintf (stderr, "out of memory!\n");
      return 0;
    }

  device->data = malloc (3 * width * height);
  if (device->data == 0)
    {
      fprintf (stderr, "out of memory!\n");
      free (device);
      return 0;
    }

  device->f = fopen (name, "w");
  if (device->f == 0)
    {
      fprintf (stderr, "couldn't open file %s\n", name);
      free (device->data);
      free (device);
      return 0;
    }
  
  device->dev.width = width;
  device->dev.height = height;
  device->dev.scale = scale;
  device->dev.gamma = gamma;

  device->dev.pixel = pixel_rgb;
  device->dev.wait = wait_rgb;
  device->dev.close = close_rgb;

  

  return (Image_out_device *)device;
}
