// x.c
// (C) 1998 lars brinkhoff

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>

#include "image_out_device.h"

#define FLUSH_COUNT 32

#define plot(device, gc, x, y) \
	  XDrawPoint (device->display, \
		      device->window, \
		      gc, \
		      x, y)

struct device
{
  Image_out_device dev;
  Display *display;
  int screen;
  int depth;
  Visual *visual;
  Window window;
  GC gc;
  GC gc2;
  GC colors[256];
  Pixmap image;
  int flush_counter;
  XWindowAttributes attr;
};

static const int D[4][4] =
  { {  1,  9,  3, 12 },
    { 15,  5, 13,  7 },
    {  4, 11,  2, 10 },
    { 14,  8, 16,  6 } };

static inline int
round (double x)
{
  return (int)(x + 0.5);
}

static void
pixel_x (Image_out_device *device_, int x, int y, double r, double g, double b)
{
  struct device *device = (struct device *)device_;
  double i;
  int u, v;
  int col;
  GC gc;
  
  i = pow (0.3 * r + 0.6 * g + 0.1 * b, device->dev.gamma);

  for (u = 0; u < device->dev.scale; u++)
    {
      for (v = 0; v < device->dev.scale; v++)
	{
	  int X, Y;
	  
	  X = device->dev.scale * x + u;
	  Y = device->dev.scale * y + v;

	  if (device->depth == 1)
	    {
	      //col = i < ((double)rand () / RAND_MAX) ? 0 : 1;
	      //col = 17 * i > D[X % 4][Y % 4] ? 1 : 0;
	      col = i + (double)rand () / RAND_MAX / 17.0 - 0.5 / 17 >
		    D[X % 4][Y % 4] / 17.0 ? 1 : 0;
	      gc = device->colors[col];
	    }
	  else if (device->depth <= 8)
	    {
	      col = 255 * i + .5;
	      gc = device->colors[col];
	    }
	  else
	    {
	      XGCValues values;
	      XColor a, b;
	      char tmpcol[100];

	      sprintf (tmpcol, "#%02X%02X%02X", round (255 * r),
		                                round (255 * g),
		                                round (255 * g));
	      XAllocNamedColor (device->display, device->attr.colormap, tmpcol, &a, &b);
	      values.foreground = a.pixel;
	      XChangeGC (device->display,
			 device->gc2,
			 GCForeground,
			 &values);
	      gc = device->gc2;
	    }
	  XDrawPoint (device->display,
		      device->window,
		      gc,
		      X, Y);
	}
    }
  
  if (device->flush_counter == 0)
    {
      XFlush (device->display);
      device->flush_counter = FLUSH_COUNT;
    }

  device->flush_counter--;

#if 0
  if (device->interleave > 1 &&
      x % device->interleave == device->interleave - 1)
    {
      if (y % device->interleave == device->interleave - 1)
	{
	  int x1 = x - device->interleave + 1;
	  int y1 = y - device->interleave + 1;
	  XCopyArea (device->display,
		     device->image,
		     device->window,
		     device->gc,
		     device->dev.scale * x1, device->dev.scale * y1,
		     device->interleave * device->scale,
		     device->interleave * device->dev.scale,
		     device->dev.scale * x1, device->dev.scale * y1);
	  XFlush (device->display);
	}
    }
  else if (x + 1 == device->dev.width)
    {
      XCopyArea (device->display,
		 device->image,
		 device->window,
		 device->gc,
		 0, device->dev.scale * y,
		 device->dev.dev.width * device->dev.scale,
		 device->dev.scale,
		 0, device->dev.scale * y);
      XFlush (device->display);
    }
#endif
}

static void
wait_x (Image_out_device *device_)
{
  struct device *device = (struct device *)device_;
  XEvent evnt;
  
  XFlush (device->display);

  XSelectInput(device->display, device->window,
	       KeyPressMask | KeyReleaseMask | ExposureMask);

  for (;;)
    {
      XCopyArea (device->display,
		 device->image,
		 device->window,
		 device->gc,
		 0, 0,
		 device->dev.width * device->dev.scale,
		 device->dev.height * device->dev.scale,
		 0, 0);
      XFlush (device->display);
      XNextEvent (device->display, &evnt);
      switch(evnt.type)
	{
	case KeyPress:
	  return;
	}
    }
}

static void
close_x (Image_out_device *device_)
{
  struct device *device = (struct device *)device_;

  XDestroyWindow (device->display, device->window);
  XFlush (device->display);
}

Image_out_device *
create_x_image_out_device (char *name, int width, int height, int scale,
			   double gamma)
{
  struct device *device;
  XGCValues values;
  char tmpcol[20];
  XColor a, b;
  int i;
  
  device = malloc (sizeof (struct device));
  if (device == 0)
    return 0;
  
  device->dev.width = width;
  device->dev.height = height;
  device->dev.scale = scale;
  device->dev.gamma = gamma;
  device->flush_counter = FLUSH_COUNT;

  device->dev.pixel = pixel_x;
  device->dev.wait = wait_x;
  device->dev.close = close_x;

  device->display = XOpenDisplay (name);
  if (device->display == 0)
    return 0;

  device->screen = DefaultScreen (device->display);
  device->depth = DefaultDepth (device->display,
				device->screen);

  device->visual = DefaultVisual (device->display,
				  device->screen);
  device->window = XCreateSimpleWindow (device->display,
					RootWindow (device->display,
						    device->screen),
					0, 0,
					width * scale,
					height * scale,
					4, 0, 0);
  
  values.foreground = 0;
  values.plane_mask = 0x00ff;
  values.graphics_exposures = 0;
  device->gc = XCreateGC (device->display,
			  device->window,
			  GCForeground | GCPlaneMask |
			  GCGraphicsExposures, 
			  &values);
  
  XMapWindow (device->display, device->window);
  
  device->image = XCreatePixmap (device->display,
				 device->window,
				 width * scale,
				 height * scale,
				 device->depth);
  
  XGetWindowAttributes (device->display,
			device->window, &device->attr);
  
  if (device->depth <= 8)
    {
      for (i= 0; i < 1 << device->depth; i++)
	{
	  int z;
	  
	  z = 256 >> device->depth;
	  sprintf (tmpcol, "#%02X%02X%02X", i * z, i * z, i * z);
	  XAllocNamedColor (device->display, device->attr.colormap,
			    tmpcol, &a, &b);
	  values.foreground = a.pixel;
	  device->colors[i] = XCreateGC (device->display,
					 device->window,
					 GCForeground, &values);
	}
    }

  XAllocNamedColor (device->display, device->attr.colormap, "#000000", &a, &b);
  values.foreground = a.pixel;
  device->gc2 = XCreateGC (device->display,
			   device->window,
			   GCForeground, &values);

  return (Image_out_device *)device;
}

#if 0
int main (int argc, char *argv[])
{
  Image_out_device *device;
  int x, y;

  device = create_x_image_out_device (0, 100, 100, 4, 1);
  
  for (y = 0; y < 100; y++)
    for (x = 0; x < 100; x++)
      device->pixel (device, x, y, 0, y / 100.0, 0);

  device->wait (device);
  device->close (device);

  return 0;
}
#endif
